#define EXTERN extern
#include "pdftexd.h"
#include "webmacros.h"

#ifdef WIN32
#define M_PI       3.1415926535897932385E0  /*Hex  2^ 1 * 1.921FB54442D18 */
#define M_PI_2     1.5707963267948966192E0  /*Hex  2^ 0 * 1.921FB54442D18 */
#define M_PI_4     7.8539816339744830962E-1 /*Hex  2^-1 * 1.921FB54442D18 */
#endif

#define MAX_CHAR_CODE       255
#define MOVE_CHARS_OFFSET   160
#define PRINTF_BUF_SIZE     1024

#define check_buf(size, buf_size)                         \
    if ((size) >= buf_size - 2)                           \
        pdftex_fail("buffer overflow [%i bytes]", (int)(buf_size))

#define append_char_to_buf(c, p, buf, buf_size) do {       \
    if (c == 9)                                            \
        c = 32;                                            \
    if (c == 13 || c == EOF)                               \
        c = 10;                                            \
    if (c != ' ' || (p > buf && p[-1] != 32)) {            \
        check_buf(p - buf, buf_size);                      \
        *p++ = c;                                          \
    }                                                      \
} while (0)

#define append_eol(p, buf, buf_size) do {                  \
    if (p - buf > 1 && p[-1] != 10) {                      \
        check_buf(p - buf, buf_size);                      \
        *p++ = 10;                                         \
    }                                                      \
    if (p - buf > 2 && p[-2] == 32) {                      \
        p[-2] = 10;                                        \
        p--;                                               \
    }                                                      \
    *p = 0;                                                \
} while (0)

#define remove_eol(p, buf) do {                            \
    p = strend(buf) - 1;                                   \
    if (*p == 10)                                          \
        *p = 0;                                            \
} while (0)

#define skip(p, c)   if (*p == c)  p++

#define entry_room(t, s) do {                              \
    if (t##_tab == 0) {                                    \
        t##_max = (s);                                     \
        t##_tab = xtalloc(t##_max, t##_entry);             \
        t##_ptr = t##_tab;                                 \
    }                                                      \
    else if (t##_ptr - t##_tab == t##_max) {               \
        t##_tab = xretalloc(t##_tab, t##_max + (s), t##_entry); \
        t##_ptr = t##_tab + t##_max;                       \
        t##_max += (s);                                    \
    }                                                      \
} while (0)

#define xfree(p)            do { if (p != 0) free(p); p = 0; } while (0)
#define strend(s)           strchr(s, 0)
#define xtalloc             XTALLOC
#define xretalloc           XRETALLOC

#define ASCENT_CODE         0
#define CAPHEIGHT_CODE      1
#define DESCENT_CODE        2
#define FONTNAME_CODE       3
#define ITALIC_ANGLE_CODE   4
#define STEMV_CODE          5
#define XHEIGHT_CODE        6
#define FONTBBOX1_CODE      7
#define FONTBBOX2_CODE      8
#define FONTBBOX3_CODE      9
#define FONTBBOX4_CODE      10
#define MAX_KEY_CODE        (FONTBBOX1_CODE + 1)
#define FONT_KEYS_NUM       (FONTBBOX4_CODE + 1)

#define F_INCLUDED          0x01
#define F_SUBSETTED         0x02
#define F_TRUETYPE          0x04
#define F_BASEFONT          0x08
#define F_NOPARSING         0x10
#define F_PGCFONT           0x20

#define set_included(fm)    ((fm)->type |= F_INCLUDED)
#define set_subsetted(fm)   ((fm)->type |= F_SUBSETTED)
#define set_truetype(fm)    ((fm)->type |= F_TRUETYPE)
#define set_basefont(fm)    ((fm)->type |= F_BASEFONT)
#define set_noparsing(fm)   ((fm)->type |= F_NOPARSING)
#define set_pcgfont(fm)     ((fm)->type |= F_PGCFONT)

#define unset_included(fm)  ((fm)->type &= ~F_INCLUDED)
#define unset_subsetted(fm) ((fm)->type &= ~F_SUBSETTED)
#define unset_truetype(fm)  ((fm)->type &= ~F_TRUETYPE)
#define unset_basefont(fm)  ((fm)->type &= ~F_BASEFONT)
#define unset_noparsing(fm) ((fm)->type &= ~F_NOPARSING)
#define unset_pcgfont(fm)   ((fm)->type &= ~F_PGCFONT)

#define is_included(fm)     ((fm)->type & F_INCLUDED)
#define is_subsetted(fm)    ((fm)->type & F_SUBSETTED)
#define is_truetype(fm)     ((fm)->type & F_TRUETYPE)
#define is_basefont(fm)     ((fm)->type & F_BASEFONT)
#define is_noparsing(fm)    ((fm)->type & F_NOPARSING)

#define is_reencoded(fm)    ((fm)->encoding >= 0)
#define is_slanted(fm)      ((fm)->slant != 0)
#define is_extended(fm)     ((fm)->extend != 0)
#define is_fontfile(fm)     ((fm)->ff_name != 0)

#define is_t1fontfile(fm)   (is_fontfile(fm) && !is_truetype(fm))
#define is_pcgfont(fm)      (!is_basefont(fm) && !is_fontfile(fm))
#define need_encoding_obj(fm) (is_reencoded(fm) || is_subsetted(fm))
    /*
    (is_reencoded(fm) && !is_subsetted(fm) && !is_truetype(fm))
    */

#define unset_fontfile(fm)  xfree((fm)->ff_name)

typedef struct {
    char *pdfname;
    char *t1name;
    union {
      integer i;
      char *s;
    } value;
    boolean valid;
} key_entry;

typedef struct {
    boolean loaded;     /* encoding is already loaded? */
    boolean updated;    /* encoding is already updated for 0--31 chars? */
    internalfontnumber firstfont;   /* first font that has accessed to this
                                       encoding; move chars 0--31 to higher
                                       area according to this font */
    char *name;         /* encoding file name */
    integer objnum;     /* object number */
    char **glyph_names;
} enc_entry;

typedef struct {
    char *tfm_name;         /* TFM file name */
    char *base_name;        /* PostScript name */
    integer flags;          /* font flags */
    char *ff_name;          /* font file name */
    char *ex_name;          /* expanded font file name */
    char *subset_tag;       /* pseudoUniqueTag for subsetted font */
    int encoding;           /* index to table of encoding vectors */
    internalfontnumber texfont; /* inverse of pdffontmap */
    int type;               /* font type (T1/TTF/...) */
    integer slant;          /* SlantFont */
    integer extend;         /* ExtendFont */
    boolean found;          /* to avoid researching fonts that have been found */
    integer ff_objnum;      /* font file object number */
    integer fn_objnum;      /* font name object number */
    char *charset;          /* string containing used glyphs */
} fm_entry;

/*
 *   Copied from pkin.h
 *
 *   Type declarations.  integer must be a 32-bit signed; shalfword must
 *   be a sixteen-bit signed; halfword must be a sixteen-bit unsigned;
 *   quarterword must be an eight-bit unsigned.
 *
 *   A chardesc describes an individual character. 
 */

typedef short shalfword ;
typedef short Boolean ;
typedef struct {
   integer charcode,
           cwidth, 
           cheight,
           xoff, 
           yoff, 
           xescape;
   halfword *raster;
} chardesc;

extern fm_entry *fm_cur, *fm_ptr, *fm_tab;
extern enc_entry *enc_ptr, *enc_tab;
extern char *filename;
extern char *mapfiles;
extern internalfontnumber tex_font;
extern integer t1_length1, t1_length2, t1_length3;
extern char **t1_glyph_names, *t1_builtin_glyph_names[];
extern integer ttf_length;
extern char notdef[];
extern char nontfm[];
extern boolean fontfile_found;
extern key_entry font_keys[];

extern int xgetc(FILE *);
extern int xputc(int, FILE *);
extern size_t xfwrite(void *, size_t, size_t, FILE *);
extern int xfflush(FILE *);
extern void pdftex_fail(char *, ...);
extern void pdftex_warn(char *, ...);
extern void pdf_printf(char *, ...);
extern void pdf_puts(char *);
extern void tex_printf(char *, ...);
extern strnumber maketexstring(char *);
extern char *makecstring(integer);
extern boolean str_eq_cstr(strnumber, char *);
extern void fm_read_info(void);
extern fm_entry *fm_ext_entry(internalfontnumber);
extern integer myatol(char **);
extern integer myatodim(char **);
extern boolean t1_read_enc(fm_entry *);
extern void writet1(void);
extern void writettf(void);
extern void writet3(int, internalfontnumber);
extern int readchar(boolean , chardesc *cd);
extern integer add_enc(char *);
extern boolean get_enc(fm_entry *);
extern void read_enc(integer);
extern void write_enc(char **, integer);
extern void update_enc(internalfontnumber, char**);
extern void enc_free(void);
extern void epdf_free(void);
extern void fm_free(void);
extern void img_free(void);
extern void vf_free(void);

